(* Theorems about CPS transformation *)

Require Import Coq.Lists.List.
Require Import ListFacts.
Require Import "Variable".
Require Import Tactics.
Require Import SourceLang.
Require Import CpsLang.
Require Import Cps.


Definition scctxt (Cs : sctxt) (Cc : cctxt) (vars : list var) :=
  forall (x : var) (T : stype),
  In x vars ->
  in' x T Cs ->
  in' x (cps_type T) Cc.


Ltac intro_fresh x :=
  match goal with
  | |- context[fresh ?l] =>
      set (x := fresh l)
  end.
Ltac fresh_not_eq :=
  apply fresh_not_eq; auto_in.


Lemma scctxt_weakening_l :
  forall (Cs : sctxt) (Cc : cctxt) (l1 l2 : list var),
  incl l2 l1 ->
  scctxt Cs Cc l1 ->
  scctxt Cs Cc l2.
Proof.
  firstorder.
Qed.


Theorem cps_term_preserves_typing :
  forall (Cs : sctxt) (T : stype) (t : sterm),
  sterm_typing Cs t T ->
  forall Cc : cctxt,
  scctxt Cs Cc (fvars_in_sterm t) ->
  cterm_typing Cc (cps_term t) (! ! cps_type T).
Proof.
  induction 1 as [ C x T H1
                 | C
                 | C x TA TR t H1 IH1
                 | C TA TR tf ta H1 IH1 H2 IH2 ];
  simpl;
  intros Cc HA.

  (* case stt_fvar *)
  intro_fresh xk.
  apply ctt_abs.
  apply ctt_app with (TA := cps_type T).
    apply ctt_fvar.
    left; split; reflexivity.

    apply ctt_fvar.
    right; split.
      fresh_not_eq.
      apply HA; auto_in.

  (* case stt_unit *)
  intro_fresh xk.
  apply ctt_abs.
  apply ctt_app with (TA := cps_type sy_unit).
    apply ctt_fvar.
    left; split; reflexivity.

    apply ctt_unit.

  (* case stt_abs *)
  intro_fresh xk.
  apply ctt_abs.
  apply ctt_app with (TA := cps_type (sy_func TA TR)).
    apply ctt_fvar.
    left; split; reflexivity.

    apply ctt_abs; fold cps_type.
    apply IH1.
    intros x2 T HB [[<- <-] | [HC HD]]; fold (@in' var stype) in *.
      left; split; reflexivity.

      right; split; [ exact HC | ].
      right; split; [ apply fresh_not_eq | apply HA, HD ];
      apply neq_remove_In, HB;
      contradict HC; symmetry; exact HC.

  (* case stt_app *)
  intro_fresh xk. intro_fresh xf. intro_fresh xa.
  apply ctt_abs.
  set (Cck := (xk, !cps_type TR) :: Cc).
  assert (HAk : scctxt C Cck (fvars_in_sterm tf ++ fvars_in_sterm ta)).
    intros x T HB HC.
    right; split.
      fresh_not_eq.
      apply HA; assumption.
  apply ctt_app with (TA := ! cps_type (sy_func TA TR)).
    apply IH1.
    refine (scctxt_weakening_l _ _ _ _ _ HAk); auto_in.

    apply ctt_abs.
    set (Ccf := (xf, cps_type (sy_func TA TR)) :: Cck).
    assert (HAf : scctxt C Ccf ((*fvars_in_sterm tf ++*) fvars_in_sterm ta)).
      intros x T HB HC.
      right; split.
        fresh_not_eq.
        apply HAk; auto_in.
    apply ctt_app with (TA := ! cps_type TA).
      apply IH2, HAf.

      apply ctt_abs.
      set (Cca := (xa, cps_type TA) :: Ccf).
      apply ctt_app with (TA := ! cps_type TR);
      [ apply ctt_app with (TA := cps_type TA) | ];
      apply ctt_fvar;
      repeat ((left; split; reflexivity)
          || (right; split; [ fresh_not_eq | ])).
Qed.
