(* Definition of the CPS-transformed language *)

Require Import Coq.Lists.List.
Require Import ListFacts.
Require Import "Variable".


Section Terms.

(* term *)
Inductive cterm : Set :=
  | ct_fvar (x : var)
  | ct_unit
  | ct_abs (x : var) (t : cterm)
  | ct_app (tf ta : cterm).

Fixpoint fvars_in_cterm (t : cterm) : list var :=
  match t with
  | ct_fvar x =>
      x :: nil
  | ct_unit =>
      nil
  | ct_abs x t1 =>
      remove eq_var_dec x (fvars_in_cterm t1)
  | ct_app tf ta =>
      fvars_in_cterm tf ++ fvars_in_cterm ta
  end.

End Terms.


Section Types.

(* type *)
Inductive ctype : Set :=
  | cy_unit
  | cy_func (TA TR : ctype)
  | cy_bottom.

End Types.


Section Contexts.

(* typing context *)
Definition cctxt : Set :=
  list (var * ctype).

Definition domain_of_cctxt : cctxt -> list var :=
  map (@fst _ _).

End Contexts.


Section Rules.

(* term typing rule *)
Inductive cterm_typing : cctxt -> cterm -> ctype -> Prop :=
  | ctt_fvar :
      forall (C : cctxt) (x : var) (T : ctype),
      in' x T C ->
      cterm_typing C (ct_fvar x) T
  | ctt_unit :
      forall C : cctxt,
      cterm_typing C ct_unit cy_unit
  | ctt_abs :
      forall (C : cctxt) (x : var) (TA TR : ctype) (t : cterm),
      cterm_typing ((x, TA) :: C) t TR ->
      cterm_typing C (ct_abs x t) (cy_func TA TR)
  | ctt_app :
      forall (C : cctxt) (TA TR : ctype) (tf ta : cterm),
      cterm_typing C tf (cy_func TA TR) ->
      cterm_typing C ta TA ->
      cterm_typing C (ct_app tf ta) TR.

End Rules.


Bind Scope ctype_scope with ctype.
Delimit Scope ctype_scope with ctype.
Notation "! T" := (cy_func T cy_bottom)
  (at level 35) : ctype_scope.
