(* Definition of variables used in syntax definitions *)

Require Import Coq.Logic.Decidable.
Require Import Coq.Arith.Le.
Require Import Coq.Arith.Lt.
Require Import Coq.Arith.Peano_dec.
Require Import Coq.Arith.Compare_dec.
Require Import Coq.Lists.List.


Definition var := nat.

Definition eq_var_dec :
  forall x1 x2 : var, {x1 = x2} + {x1 <> x2} := eq_nat_dec.
Definition dec_eq_var :
  forall x1 x2 : var, decidable (x1 = x2) := dec_eq_nat.

(* returns a new variable that is not included in the list *)
Definition fresh' (l : list var) :
  {x : var | forall x' : var, In x' l -> x' < x}.
Proof.
  induction l as [ | x l' [ x' IH ] ].
    exists O; intros x' H; case H.

    destruct (le_lt_dec x' x) as [ H | H ].
      exists (S x); intros x'' Hin; apply le_n_S.
      inversion Hin as [ Heq | Heq ].
        rewrite Heq; apply le_n.

        apply le_S_n, le_S, (lt_le_trans x'' x' x), H.
        apply IH, Heq.

      exists x'; intros x'' Hin.
      inversion Hin as [ Heq | Heq ].
        rewrite Heq in H; apply H.

        apply IH, Heq.
Defined.

Definition fresh (l : list var) : var := proj1_sig (fresh' l).

Lemma fresh_not_in :
  forall (l : list var), ~ In (fresh l) l.
Proof.
  intros l H; unfold fresh, proj1_sig in H.
  destruct (fresh' l) as [ x H' ] in H.
  apply H' in H; apply (lt_irrefl _ H).
Qed.

Lemma fresh_not_eq :
  forall (x : var) (l : list var),
  In x l -> fresh l <> x.
Proof.
  intros x l H1 H2.
  apply (fresh_not_in l).
  rewrite H2; assumption.
Qed.
